module Main exposing (..)

import Browser
import Html exposing (..)
import Html.Attributes exposing (..)
import Html.Events exposing (..)
import Svg exposing (svg)
import Svg.Attributes
import Character exposing (Character)
import Stats exposing (Stat)
import BoundedInt exposing (BoundedInt)
import Stats exposing (Category)
import Creation exposing (CreationPoints)
import OrderedCategories as OCat exposing (OCat, CategoriesType)

main : Program () Model Msg
main = 
    Browser.document
        { init = init
        , view = view
        , update = update
        , subscriptions = subscriptions
        }

init : () -> ( Model, Cmd Msg )
init _ =
    (   { character = Character.new
        , modal = Nothing
        , modalValue = ""
        }
    , Cmd.none
    )

type alias Model = 
    { character : Character
    , modal : Maybe ModalType
    , modalValue : String
    }

type Msg 
    = OpenModal ModalType
    | CloseModal
    | ChangeModalValue String
    | SaveModalChange ModalType
    | ChangeStat Stat
    | ChangePriority CategoriesType Category Int

type ModalType
    = Name

update : Msg -> Model -> ( Model, Cmd Msg )
update msg model = 
    case msg of
        OpenModal modalType -> 
            (   { model 
                | modal = ( Just modalType ) 
                , modalValue = modalValue modalType model.character
                }
            , Cmd.none
            )
        CloseModal ->
            ( { model | modal = Nothing }
            , Cmd.none
            )
        ChangeModalValue value ->
            ( { model | modalValue = value }
            , Cmd.none
            )
        SaveModalChange modalType ->
            case modalType of
                Name ->
                    (   { model 
                        | character = Character.changeName model.character model.modalValue 
                        , modal = Nothing
                        }
                    , Cmd.none
                    )
        ChangeStat stat ->
            (   { model 
                | character = Character.changeStat model.character stat 
                }
            , Cmd.none
            )
        ChangePriority catType category value ->
            (   { model 
                | character = Character.changeOrder model.character catType category value
                }
            , Cmd.none
            )

view : Model -> Browser.Document Msg
view model =
    { title = "Character Sheet"
    , body =
        [ div [ class "container" ] 
            [ viewEditableText Name model.character.name
            , viewAttributes model.character.creationPoints.attributes model.character.stats
            , viewAbilities model.character.creationPoints.abilities model.character.stats
            , viewSpheres model.character.creationPoints ( Stats.spheresInList model.character.stats )
            , viewArete ( Stats.areteInList model.character.stats )
            , viewModal model
            ]
        ]
    }

viewEditableText : ModalType -> String -> Html Msg
viewEditableText modalType content =
    div []
        [ span [ class "icon-text" ]
            [ span [] [ text content ]
            , span 
                [ class "icon" 
                ] 
                [ i 
                    [ class "fas fa-pen"
                    , onClick ( OpenModal modalType ) 
                    ] []
                ]
            ]
        ]

viewAttributes : OCat -> List Stat -> Html Msg
viewAttributes oCat attributes =
    div [ class "has-text-centered" ]
        [ p [] [ text "Attributes" ]
        , div [ class "columns is-centered is-mobile" ]
            [ viewAttributeCol oCat "Physical" Stats.Physical attributes
            , viewAttributeCol oCat "Social" Stats.Social attributes
            , viewAttributeCol oCat "Mental" Stats.Mental attributes
            ]
        ]

viewAttributeCol : OCat -> String -> Category -> List Stat -> Html Msg
viewAttributeCol oCat name category attributes =
    let 
        categoryAttributes : List Stat
        categoryAttributes = Stats.attributesOfCategoryInList category attributes

        index : Int
        index = OCat.indexOf oCat.order category
        basePoints : Int
        basePoints = OCat.categoryValue oCat category
        points : Int
        points = basePoints - ( Stats.count categoryAttributes ) + ( List.length categoryAttributes )
    in
        div [ class "column is-one-fifth" ]
            (   [ p [] [ text name ]
                , creationInfo points
                , div [ class "field has-addons is-fullwidth" ]
                    [ p [ class "control is-expanded" ]
                        [ button 
                            [ class ( "button is-small is-fullwidth" ++ if index == 0 then " is-primary" else "" )
                            , onClick ( ChangePriority OCat.Attribute category 0 )
                            ] [ text "I" ]
                        ]
                    ,  p [ class "control is-expanded" ]
                        [ button 
                            [ class ( "button is-small is-fullwidth" ++ if index == 1 then " is-primary" else "" )
                            , onClick ( ChangePriority OCat.Attribute category 1 )
                            ] [ text "II" ]
                        ]
                    ,  p [ class "control is-expanded" ]
                        [ button 
                            [ class ( "button is-small is-fullwidth" ++ if index == 2 then " is-primary" else "" )
                            , onClick ( ChangePriority OCat.Attribute category 2 )
                            ] [ text "III" ]
                        ]
                    ]
                ]
                ++ 
                ( List.map viewStat categoryAttributes
                )
            )

viewAbilities : OCat -> List Stat -> Html Msg
viewAbilities oCat abilities =
    div [ class "has-text-centered" ]
        [ p [] [ text "Abilities" ]
        , div [ class "columns is-centered is-desktop" ]
            [ viewAbilityCol oCat "Talents" Stats.Talent abilities
            , viewAbilityCol oCat "Skills" Stats.Skill abilities
            , viewAbilityCol oCat "Knowledges" Stats.Knowledge abilities
            ]
        ]

viewAbilityCol : OCat -> String -> Category -> List Stat -> Html Msg
viewAbilityCol oCat name category abilities =
    let 
        categoryAbilities : List Stat
        categoryAbilities = Stats.abilitiesOfCategoryInList category abilities

        index : Int
        index = OCat.indexOf oCat.order category
        basePoints : Int
        basePoints = OCat.categoryValue oCat category
        points : Int
        points = basePoints - ( Stats.count categoryAbilities )
    in
        div [ class "column" ]
            (   [ p [] [ text name ]
                , creationInfo points
                , div [ class "field has-addons is-desktop" ]
                    [ p [ class "control is-expanded" ]
                        [ button 
                            [ class ( "button is-small is-fullwidth" ++ if index == 0 then " is-primary" else "" )
                            , onClick ( ChangePriority OCat.Ability category 0 )
                            ] [ text "I" ]
                        ]
                    ,  p [ class "control is-expanded" ]
                        [ button 
                            [ class ( "button is-small is-fullwidth" ++ if index == 1 then " is-primary" else "" )
                            , onClick ( ChangePriority OCat.Ability category 1 )
                            ] [ text "II" ]
                        ]
                    ,  p [ class "control is-expanded" ]
                        [ button 
                            [ class ( "button is-small is-fullwidth" ++ if index == 2 then " is-primary" else "" )
                            , onClick ( ChangePriority OCat.Ability category 2 )
                            ] [ text "III" ]
                        ]
                    ]
                , div [ class "fixed-grid has-6-cols" ]
                    [ div [ class "grid" ]
                        ( List.map ( viewStatInGrid ( List.length categoryAbilities ) ) ( List.indexedMap Tuple.pair categoryAbilities )
                        )
                    ]
                ]
            )

viewStatInGrid : Int -> ( Int, Stat ) -> Html Msg
viewStatInGrid totalCount indexedStat =
    let
        index : Int
        index = Tuple.first indexedStat
        stat : Stat
        stat = Tuple.second indexedStat

        offset : Int
        offset = Basics.remainderBy 3 totalCount
        dif : Int
        dif = totalCount - index

        offsetString : String
        offsetString = 
            if dif == offset
            then " is-col-start-" ++ ( String.fromInt ( 4 - dif ) )
            else ""
    in
        div [ class ( "cell is-col-span-2" ++ offsetString ) ]
            [ div [ class "notification" ]
                [ p [ class "block" ] [ text stat.name ]
                , p [ class "block" ] [ viewDots stat ]
                ]
            ]


viewSpheres : CreationPoints -> List Stat -> Html Msg
viewSpheres cp spheres =
    let
        basePoints : Int
        basePoints = cp.spheres
        points : Int
        points = basePoints - ( Stats.count spheres )
    in
        div [ class "has-text-centered" ]
            [ p [] [ text "Spheres" ]
            , div [ class "columns is-centered is-mobile" ]
                [ div [ class "column is-three-fifths" ]
                    [ creationInfo points
                    ]
                ]
            , div [ class "columns is-centered is-mobile" ]
                [ div [ class "column is-one-fifth" ] 
                    ( List.map viewStat ( Stats.everyThirdStat spheres 0 )
                    )
                , div [ class "column is-one-fifth" ] 
                    ( List.map viewStat ( Stats.everyThirdStat spheres 1 )
                    )
                , div [ class "column is-one-fifth" ] 
                    ( List.map viewStat ( Stats.everyThirdStat spheres 2 )
                    )
                ]
            ]

creationInfo : Int -> Html Msg
creationInfo points = 
    let
        notificationColor : String
        notificationColor =
            case compare points 0 of
                GT -> "is-warning"
                EQ -> "is-success"
                LT -> "is-danger"
    in
        div [ class ( "notification " ++ notificationColor ) ]
            [ text ( ( String.fromInt points ) ++ " remaining" )
            ]

viewArete : Stat -> Html Msg
viewArete arete =
    div [ class "has-text-centered" ]
        [ div [ class "columns is-centered is-mobile" ]
            [ div [ class "column is-three-fifths" ]
                [ viewStat arete
                ]
            ]
        ]

viewStat : Stat -> Html Msg
viewStat stat =
    div [ class "notification" ]
        [ p [ class "block" ] [ text stat.name ]
        , p [ class "block" ] [ viewDots stat ]
        ]

viewDots : Stat -> Html Msg
viewDots stat =
    let
        dotSize = 12*2
        dotGap = dotSize / 2
        dotOutline = 0.05 * dotSize
        marginalIncrease = (  dotSize + dotGap ) / 2
        svgOutlineMargin = ( sqrt ( 2 * ( dotOutline ^ 2 ) ) ) / 2

        rangeList = List.range 0 ( stat.dots.max - 1 )

        isFilledList = rangeList |> List.map ( (>) stat.dots.value )
    in
        svg 
            [ Svg.Attributes.width  ( String.fromFloat ( dotSize + marginalIncrease * toFloat ( stat.dots.max - 1 ) + 2 * svgOutlineMargin ) )
            , Svg.Attributes.height ( String.fromFloat ( dotSize + marginalIncrease + 2 * svgOutlineMargin ) )
            ]
            ( rangeList
                |> List.map ( toDotPosition dotSize dotGap svgOutlineMargin )
                |> List.map3 ( viewDot stat dotSize dotOutline ) isFilledList rangeList
            )

toDotPosition : Float -> Float -> Float -> Int -> ( Float, Float )
toDotPosition size gap offset id =
    let
        marginalIncrease = ( size + gap ) / 2
    in
        ( toFloat id * marginalIncrease + offset
        , toFloat ( remainderBy 2 ( id + 1 ) ) * marginalIncrease + offset
        )

viewDot : Stat -> Float -> Float -> Bool -> Int -> ( Float, Float ) -> Html Msg
viewDot stat size strokeWidth isFilled id position =
    Svg.polygon 
        [ Svg.Attributes.points 
            (   [ ( 0.5, 0.0 )
                , ( 1.0, 0.5 )
                , ( 0.5, 1.0 )
                , ( 0.0, 0.5 )
                ]
                |> List.map ( Tuple.mapBoth   ( (*) size ) ( (*) size ) )
                |> List.map ( Tuple.mapFirst  ( (+) ( Tuple.first  position ) ) )
                |> List.map ( Tuple.mapSecond ( (+) ( Tuple.second position ) ) )
                |> List.map tupleToString
                |> List.intersperse " "
                |> String.concat
            )
        , Svg.Attributes.fill "var(--bulma-body-color)"
        , Svg.Attributes.fillOpacity ( boolAsIntString isFilled )
        , Svg.Attributes.stroke "var(--bulma-body-color)"
        , Svg.Attributes.strokeWidth ( String.fromFloat strokeWidth )
        , onClick ( ChangeStat { stat | dots = BoundedInt.changeTo stat.dots ( valueOrReset stat.dots ( id + 1 ) ) } )
        ] []

tupleToString : ( Float, Float ) -> String
tupleToString tuple =
    let
        first  = String.fromFloat ( Tuple.first  tuple )
        second = String.fromFloat ( Tuple.second tuple )
    in
        first ++ "," ++ second

boolAsIntString : Bool -> String
boolAsIntString bool =
    if bool
    then "1"
    else "0"

valueOrReset : BoundedInt -> Int -> Int
valueOrReset dots value =
    if dots.value == value
    then dots.min
    else value

viewModal : Model -> Html Msg
viewModal model =
    case model.modal of
        Nothing -> div [ class "modal" ] []
        Just modalType ->
            div [ class "modal is-active" ]
                [ div 
                    [ class "modal-background" 
                    , onClick CloseModal
                    ] []
                , div [ class "modal-card" ] 
                    [ header [ class "modal-card-head" ]
                        [ p [ class "modal-card-title" ] [ text ( "Change " ++ ( modalTitle modalType ) ) ]
                        ]
                    , section [ class "modal-card-body" ] 
                        [ p [] 
                            [ input 
                                [ class "input" 
                                , type_ "text"
                                , placeholder ( "Enter New " ++ ( modalTitle modalType ) )
                                , value model.modalValue
                                , onInput ( \x -> ChangeModalValue x )
                                ] []
                            ]
                        ]
                    , footer [ class "modal-card-foot" ] 
                        [ div [ class "buttons" ] 
                            [ button 
                                [ class "button is-success"
                                , onClick ( SaveModalChange modalType )
                                ] [ text "Save Changes" ]
                            , button 
                                [ class "button"
                                , onClick CloseModal
                                ] [ text "Cancel" ]
                            ]
                        ]
                    ]
                ]

modalTitle : ModalType -> String
modalTitle modalType =
    case modalType of
        Name -> "Character Name"

modalValue : ModalType -> Character -> String
modalValue modalType character =
    case modalType of
        Name -> character.name

subscriptions : Model -> Sub Msg
subscriptions _ =
  Sub.none

-- aria functions

ariaLabel : String -> Attribute msg
ariaLabel value =
    Html.Attributes.attribute "aria-label" value

ariaHaspopup : String -> Attribute msg
ariaHaspopup value =
    Html.Attributes.attribute "aria-haspopup" value

ariaControls : String -> Attribute msg
ariaControls value =
    Html.Attributes.attribute "aria-controls" value

ariaHidden : String -> Attribute msg
ariaHidden value =
    Html.Attributes.attribute "aria-hidden" value

ariaRole : String -> Attribute msg
ariaRole value =
    Html.Attributes.attribute "role" value